package com.terraformersmc.terraform.block;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1299;
import net.minecraft.class_1750;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2680;
import net.minecraft.class_2689;
import net.minecraft.class_2741;
import net.minecraft.class_2746;
import net.minecraft.class_2758;
import net.minecraft.class_3218;
import net.minecraft.class_3481;
import java.util.Random;

/**
 * A leaves block with extended range, permitting leaves to be as far as 13 blocks away from the tree rather than the
 * limit of 6 blocks imposed by vanilla leaves. It also does not block light at all.
 */
public class ExtendedLeavesBlock extends class_2248 {
	public static final int MAX_DISTANCE = 14;
	public static final class_2746 PERSISTENT = class_2741.field_12514;
	public static final class_2758 DISTANCE = class_2758.method_11867("distance", 1, MAX_DISTANCE);

	public ExtendedLeavesBlock(class_2248.class_2251 settings) {
		super(settings);
		this.method_9590(this.method_9595().method_11664().method_11657(DISTANCE, MAX_DISTANCE).method_11657(PERSISTENT, false));
	}

	@Override
	public boolean method_9542(class_2680 state) {
		return state.method_11654(DISTANCE) == MAX_DISTANCE && !state.method_11654(PERSISTENT);
	}

	@Override
	@SuppressWarnings("deprecation")
	public void method_9514(class_2680 state, class_3218 world, class_2338 pos, Random random) {
		if (!state.method_11654(PERSISTENT) && state.method_11654(DISTANCE) == MAX_DISTANCE) {
			method_9497(state, world, pos);
			world.method_8650(pos, false);
		}

	}

	@Override
	@SuppressWarnings("deprecation")
	public void method_9588(class_2680 state, class_3218 world, class_2338 pos, Random random) {
		world.method_8652(pos, updateDistanceFromLogs(state, world, pos), 3);
	}



	@Override
	@SuppressWarnings("deprecation")
	public int method_9505(class_2680 state, class_1922 view, class_2338 pos) {
		return 0;
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_2680 method_9559(class_2680 state, class_2350 direction, class_2680 neighborState, class_1936 world, class_2338 pos, class_2338 neighborPos) {
		int distance = getDistanceFromLog(neighborState) + 1;
		if (distance != 1 || state.method_11654(DISTANCE) != distance) {
			world.method_8397().method_8676(pos, this, 1);
		}

		return state;
	}

	private static class_2680 updateDistanceFromLogs(class_2680 state, class_1936 world, class_2338 pos) {
		int distance = MAX_DISTANCE;
		class_2338.class_2340 checkPos = class_2338.class_2340.method_10109();
		Throwable caught = null;

		try {
			for (class_2350 direction : class_2350.values()) {
				// .set(pos).move(direction)
				checkPos.method_10114(pos).method_10118(direction);

				distance = Math.min(distance, getDistanceFromLog(world.method_8320(checkPos)) + 1);
				if (distance == 1) {
					break;
				}
			}
		} catch (Throwable throwable) {
			caught = throwable;
			throw throwable;
		} finally {
			if (checkPos != null) {
				if (caught != null) {
					try {
						checkPos.close();
					} catch (Throwable second) {
						caught.addSuppressed(second);
					}
				} else {
					checkPos.close();
				}
			}
		}

		return state.method_11657(DISTANCE, distance);
	}

	private static int getDistanceFromLog(class_2680 state) {
		if (class_3481.field_15475.method_15141(state.method_11614())) {
			return 0;
		} else {
			return state.method_11614() instanceof ExtendedLeavesBlock ? state.method_11654(DISTANCE) : MAX_DISTANCE;
		}
	}

	@Environment(EnvType.CLIENT)
	@Override
	public void method_9496(class_2680 state, class_1937 world, class_2338 pos, Random random) {
		class_2246.field_10503.method_9496(state, world, pos, random);
	}

	@Override
	@SuppressWarnings("deprecation")
	public boolean method_9522(class_2680 state, class_2680 neighborState, class_2350 offset) {
		// OptiLeaves optimization: Cull faces with leaf block neighbors to reduce geometry in redwood forests
		return neighborState.method_11614() instanceof ExtendedLeavesBlock;
	}

	@Override
	@SuppressWarnings("deprecation")
	public boolean method_16362(class_2680 state, class_1922 view, class_2338 pos) {
		return false;
	}

	@Override
	@SuppressWarnings("deprecation")
	public boolean method_9523(class_2680 state, class_1922 view, class_2338 pos, class_1299<?> entity) {
		return entity == class_1299.field_6081 || entity == class_1299.field_6104;
	}

	@Override
	protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
		builder.method_11667(DISTANCE, PERSISTENT);
	}

	@Override
	public class_2680 method_9605(class_1750 context) {
		return updateDistanceFromLogs(this.method_9564().method_11657(PERSISTENT, true), context.method_8045(), context.method_8037());
	}
}
