package com.terraformersmc.terraform.block;

import it.unimi.dsi.fastutil.objects.Object2IntMap;
import it.unimi.dsi.fastutil.objects.Object2IntOpenHashMap;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.block.*;
import net.minecraft.class_1268;
import net.minecraft.class_1269;
import net.minecraft.class_1657;
import net.minecraft.class_1766;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2246;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2397;
import net.minecraft.class_2498;
import net.minecraft.class_259;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_2689;
import net.minecraft.class_2746;
import net.minecraft.class_3417;
import net.minecraft.class_3419;
import net.minecraft.class_3612;
import net.minecraft.class_3726;
import net.minecraft.class_3965;
import java.util.Random;
import java.util.function.Supplier;

// Rather complex: combine the function of leaves, logs, and cobblestone walls.

/**
 * A very complex smaller log block that can connect on all 6 axes, can be waterlogged, and can have leaves embedded.
 * Used for things like the Sakura tree.
 */
public class SmallLogBlock extends BareSmallLogBlock {

	public static final class_2746 HAS_LEAVES = class_2746.method_11825("has_leaves");

	private static final int UP_MASK = 1 << class_2350.field_11036.ordinal();
	private static final int DOWN_MASK = 1 << class_2350.field_11033.ordinal();
	private static final int NORTH_MASK = 1 << class_2350.field_11043.ordinal();
	private static final int EAST_MASK = 1 << class_2350.field_11034.ordinal();
	private static final int SOUTH_MASK = 1 << class_2350.field_11035.ordinal();
	private static final int WEST_MASK = 1 << class_2350.field_11039.ordinal();

	protected final class_265[] collisionShapes;
	protected final class_265[] boundingShapes;
	private final Object2IntMap<class_2680> SHAPE_INDEX_CACHE = new Object2IntOpenHashMap<>();

	private final class_2248 leaves;
	private final Supplier<class_2248> stripped;

	public SmallLogBlock(class_2248 leaves, Supplier<class_2248> stripped, class_2251 settings) {
		super(stripped, settings);
		this.method_9590(this.method_9595().method_11664()
			.method_11657(UP, false)
			.method_11657(DOWN, false)
			.method_11657(WEST, false)
			.method_11657(EAST, false)
			.method_11657(NORTH, false)
			.method_11657(SOUTH, false)
			.method_11657(WATERLOGGED, false)
			.method_11657(HAS_LEAVES, false)
		);

		this.collisionShapes = this.createShapes(5);
		this.boundingShapes = this.createShapes(5);
		this.leaves = leaves;
		this.stripped= stripped;
	}

	@Environment(EnvType.CLIENT)
	@Override
	public void method_9496(class_2680 state, class_1937 world, class_2338 pos, Random random) {
		if (state.method_11654(HAS_LEAVES)) {
			class_2246.field_10503.method_9496(state, world, pos, random);
		}
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_1269 method_9534(class_2680 state, class_1937 world, class_2338 pos, class_1657 player, class_1268 hand, class_3965 result) {
		class_1799 held = player.method_5998(hand);

		if (held.method_7947() >= 1 && held.method_7909() == class_1792.field_8003.get(leaves) && !state.method_11654(HAS_LEAVES)) {
			if (!player.method_7337()) {
				held.method_7934(1);
			}

			class_2498 sounds = leaves.method_9564().method_11638();
			world.method_8396(player, pos, sounds.method_10598(), class_3419.field_15245, (sounds.method_10597() + 1.0F) / 2.0F, sounds.method_10599() * 0.8F);

			class_2680 previous = state;

			state = state.method_11657(HAS_LEAVES, true);

			if (state.method_11654(UP) && world.method_8320(pos.method_10084()).method_11614() instanceof class_2397) {
				state = state.method_11657(UP, false);
			}

			if (state.method_11654(DOWN) && world.method_8320(pos.method_10074()).method_11614() instanceof class_2397) {
				state = state.method_11657(DOWN, false);
			}

			if (state.method_11654(WEST) && world.method_8320(pos.method_10067()).method_11614() instanceof class_2397) {
				state = state.method_11657(WEST, false);
			}

			if (state.method_11654(EAST) && world.method_8320(pos.method_10078()).method_11614() instanceof class_2397) {
				state = state.method_11657(EAST, false);
			}

			if (state.method_11654(NORTH) && world.method_8320(pos.method_10095()).method_11614() instanceof class_2397) {
				state = state.method_11657(NORTH, false);
			}

			if (state.method_11654(SOUTH) && world.method_8320(pos.method_10072()).method_11614() instanceof class_2397) {
				state = state.method_11657(SOUTH, false);
			}

			world.method_8501(pos, method_9582(previous, state, world, pos));

			return class_1269.field_5812;
		} else if(stripped != null && held.method_7909() instanceof class_1766) {
			class_1766 tool = (class_1766) held.method_7909();

			if(tool.method_7856(state) || tool.method_7865(held, state) > 1.0F) {
				world.method_8396(player, pos, class_3417.field_14675, class_3419.field_15245, 1.0F, 1.0F);

				if(!world.field_9236) {
					class_2680 target = stripped.get().method_9564()
						.method_11657(BareSmallLogBlock.UP, state.method_11654(BareSmallLogBlock.UP))
						.method_11657(BareSmallLogBlock.DOWN, state.method_11654(BareSmallLogBlock.DOWN))
						.method_11657(BareSmallLogBlock.NORTH, state.method_11654(BareSmallLogBlock.NORTH))
						.method_11657(BareSmallLogBlock.SOUTH, state.method_11654(BareSmallLogBlock.SOUTH))
						.method_11657(BareSmallLogBlock.EAST, state.method_11654(BareSmallLogBlock.EAST))
						.method_11657(BareSmallLogBlock.WEST, state.method_11654(BareSmallLogBlock.WEST))
						.method_11657(BareSmallLogBlock.WATERLOGGED, state.method_11654(BareSmallLogBlock.WATERLOGGED))
						.method_11657(SmallLogBlock.HAS_LEAVES, state.method_11654(SmallLogBlock.HAS_LEAVES));

					world.method_8501(pos, target);

					held.method_7956(1, player, consumedPlayer -> consumedPlayer.method_20236(hand));
				}

				return class_1269.field_5812;
			}
		}

		return class_1269.field_5814;
	}

	@Override
	protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
		super.method_9515(builder);

		builder.method_11667(HAS_LEAVES);
	}

	private boolean shouldConnectTo(class_2680 state, boolean solid, boolean leaves) {
		class_2248 block = state.method_11614();

		return solid || (!leaves && block instanceof class_2397) || block instanceof BareSmallLogBlock;
	}

	public class_2680 getNeighborUpdateState(class_2680 state, class_2350 fromDirection, class_2680 neighbor, class_1936 world, class_2338 pos, class_2338 neighborPos) {
		if (state.method_11654(WATERLOGGED)) {
			world.method_8405().method_8676(pos, class_3612.field_15910, class_3612.field_15910.method_15789(world));
		}

		boolean leaves = state.method_11654(HAS_LEAVES);

		boolean up = fromDirection == class_2350.field_11036 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11033), leaves) || state.method_11654(UP);
		boolean down = fromDirection == class_2350.field_11033 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11036), leaves) || state.method_11654(DOWN);
		boolean north = fromDirection == class_2350.field_11043 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11035), leaves) || state.method_11654(NORTH);
		boolean east = fromDirection == class_2350.field_11034 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11039), leaves) || state.method_11654(EAST);
		boolean south = fromDirection == class_2350.field_11035 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11043), leaves) || state.method_11654(SOUTH);
		boolean west = fromDirection == class_2350.field_11039 && this.shouldConnectTo(neighbor, class_2248.method_20045(neighbor, world, neighborPos, class_2350.field_11034), leaves) || state.method_11654(WEST);

		return state
			.method_11657(UP, up)
			.method_11657(DOWN, down)
			.method_11657(NORTH, north)
			.method_11657(EAST, east)
			.method_11657(SOUTH, south)
			.method_11657(WEST, west);
	}

	private int getShapeIndex(class_2680 requested) {
		return this.SHAPE_INDEX_CACHE.computeIntIfAbsent(requested, state -> {
			int mask = 0;

			if (state.method_11654(UP)) {
				mask |= UP_MASK;
			}

			if (state.method_11654(DOWN)) {
				mask |= DOWN_MASK;
			}

			if (state.method_11654(NORTH)) {
				mask |= NORTH_MASK;
			}

			if (state.method_11654(EAST)) {
				mask |= EAST_MASK;
			}

			if (state.method_11654(SOUTH)) {
				mask |= SOUTH_MASK;
			}

			if (state.method_11654(WEST)) {
				mask |= WEST_MASK;
			}

			return mask;
		});
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_265 method_9530(class_2680 state, class_1922 view, class_2338 pos, class_3726 context) {
		return state.method_11654(HAS_LEAVES) ? class_259.method_1077() : this.boundingShapes[this.getShapeIndex(state)];
	}

	@Override
	@SuppressWarnings("deprecation")
	public class_265 method_9549(class_2680 state, class_1922 view, class_2338 pos, class_3726 context) {
		return state.method_11654(HAS_LEAVES) ? class_259.method_1077() : this.collisionShapes[this.getShapeIndex(state)];
	}
}
